# -*- coding: utf-8 -*-
import re

REGEX_ELEMENT_V1 = re.compile(r"""(?P<element>[a-zA-Z]{1,2})\s+Element""")
REGEX_ELEMENT_V2 = re.compile(
    r"""\s*element\s*=\s*['"]\s*(?P<element>[a-zA-Z]{1,2})\s*['"].*"""
)

PATTERN_FLOAT = r"[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?"
REGEX_Z_VALENCE_V1 = re.compile(
    r"""(?P<z_valence>""" + PATTERN_FLOAT + r""")\s+Z valence"""
)
REGEX_Z_VALENCE_V2 = re.compile(
    r"""\s*z_valence\s*=\s*['"]\s*(?P<z_valence>""" + PATTERN_FLOAT + r""")\s*['"].*"""
)

REGEX_PSEUDO_TYPE_V1 = re.compile(
    r"""\s*(?P<pseudo_type>(NC|SL|1\/r|US|USPP|PAW))\s*.*\s*pseudopotential"""
)
REGEX_PSEUDO_TYPE_V2 = re.compile(
    r"""\s*pseudo_type\s*=\s*['"]\s*(?P<pseudo_type>(NC|SL|1\/r|US|USPP|PAW))\s*['"].*"""
)


def parse(content: str):
    """parse all"""
    return {
        "element": parse_element(content),
        "pp_type": parse_pseudo_type(content),
        "z_valence": parse_z_valence(content),
    }


def parse_element(content: str):
    """Parse the content of the UPF file to determine the element.
    :param stream: a filelike object with the binary content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    for regex in [REGEX_ELEMENT_V2, REGEX_ELEMENT_V1]:
        match = regex.search(content)

        if match:
            return match.group("element")

    raise ValueError(f"could not parse the element from the UPF content: {content}")


def parse_z_valence(content: str) -> int:
    """Parse the content of the UPF file to determine the Z valence.
    :param stream: a filelike object with the binary content of the file.
    :return: the Z valence.
    """
    for regex in [REGEX_Z_VALENCE_V2, REGEX_Z_VALENCE_V1]:
        match = regex.search(content)

        if match:
            z_valence = match.group("z_valence")

            try:
                z_valence = float(z_valence)
            except ValueError as exception:
                raise ValueError(
                    f"parsed value for the Z valence `{z_valence}` is not a valid number."
                ) from exception

            if int(z_valence) != z_valence:
                raise ValueError(
                    f"parsed value for the Z valence `{z_valence}` is not an integer."
                )

            return int(z_valence)

    raise ValueError(f"could not parse the Z valence from the UPF content: {content}")


def parse_pseudo_type(content: str) -> str:
    """Parse the content of the UPF file to determien the pseudo type.
    either in one of the NC | SL | 1/r | US | PAW
    :param stream: a filelike object with the binary content of the file.
    :return: the pseudo type str.
    """
    for regex in [REGEX_PSEUDO_TYPE_V2, REGEX_PSEUDO_TYPE_V1]:
        match = regex.search(content)

        if match:
            raw_type = match.group("pseudo_type")
            if "US" in raw_type:
                return "us"
            elif "NC" in raw_type or "SL" in raw_type:
                return "nc"
            elif "PAW" in raw_type:
                return "paw"
            else:
                return raw_type

    raise ValueError(f"could not parse the pseudo_type from the UPF content: {content}")



from aiida.engine import WorkChain
from aiida.orm import Bool, CalcJobNode

from aiida_sssp_workflow.workflows.common import clean_workdir, operate_calcjobs


class SelfCleanWorkChain(WorkChain):
    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input(
            "clean_workdir",
            valid_type=Bool,
            default=lambda: Bool(True),
            help="If `True`, work directories of all called non-cached calculations will be cleaned"
            " at the end of execution, and cached calculations will invalid from cache.",
        )

    def on_terminated(self):
        """Clean the working directories of all child calculations if `clean_worchain=True` in the inputs.
        If clean, important to also invalid calcjob nodes for caching. Since otherwise there will be chance
        that not the caching node but cleaned node used for caching in bands and phonon that will lead to the
        parent_folder empty issue.
        In big verification workflow, only _caching workflow is not process clean step at last but purge remote
        folder after the whole verfication workflow finished.
        hardcode the invalid_caching.
        """
        super().on_terminated()

        if not self.inputs.clean_workdir.value:
            self.report(f"{type(self)}: remote folders will not be cleaned")
            return

        # The clen_wordir is modified to not clean the remote folder of the caching node.
        cleaned_calcs = operate_calcjobs(
            self.node, operator=clean_workdir, all_same_nodes=False
        )

        if cleaned_calcs:
            self.report(
                f"cleaned remote folders of calculations: {' '.join(map(str, cleaned_calcs))}"
            )

        # This is not turned on since it will make all finished workflow not cacheable.
        # I need to find a way to properly work around the parent_folder empty issue.
        # cache_invalid_calcs = operate_calcjobs(
        #     self.node, operator=invalid_cache, all_same_nodes=False
        # )

        # if cache_invalid_calcs:
        #     self.report(
        #         f"Invalid cache of cached calculations: {' '.join(map(str, cache_invalid_calcs))}"
        #     )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""CLI to inspect the results of the workflow"""
import json

import click
import matplotlib.pyplot as plt
import numpy as np
from aiida import orm

from aiida_sssp_workflow.cli import cmd_root


def birch_murnaghan(V, E0, V0, B0, B01):
    """
    Return the energy for given volume (V - it can be a vector) according to
    the Birch Murnaghan function with parameters E0,V0,B0,B01.
    """
    r = (V0 / V) ** (2.0 / 3.0)
    return E0 + 9.0 / 16.0 * B0 * V0 * (
        (r - 1.0) ** 3 * B01 + (r - 1.0) ** 2 * (6.0 - 4.0 * r)
    )


def eos_plot(
    ax, ref_ae_data, line_data, energy0, volumes, energies, nu, title="EOS", fontsize=8
):
    """plot EOS result on ax"""
    dense_volume_max = max(volumes)
    dense_volume_min = min(volumes)

    dense_volumes = np.linspace(dense_volume_min, dense_volume_max, 100)

    ref_V0, ref_B0, ref_B01 = ref_ae_data
    V0, B0, B01 = line_data

    ae_eos_fit_energy = birch_murnaghan(
        V=dense_volumes,
        E0=energy0,  # in future update E0 from referece json, where ACWF has E0 stored.
        V0=ref_V0,
        B0=ref_B0,
        B01=ref_B01,
    )
    psp_eos_fit_energy = birch_murnaghan(
        V=dense_volumes,
        E0=energy0,
        V0=V0,
        B0=B0,
        B01=B01,
    )

    # Plot EOS
    ax.tick_params(axis="y", labelsize=6, rotation=45)

    ax.plot(volumes, energies, "ob", label="RAW equation of state")
    ax.plot(dense_volumes, ae_eos_fit_energy, "--r", label="AE")
    ax.axvline(V0, linestyle="--", color="gray")

    ax.plot(dense_volumes, psp_eos_fit_energy, "-b", label=f"Pseudo")
    ax.fill_between(
        dense_volumes,
        ae_eos_fit_energy,
        psp_eos_fit_energy,
        alpha=0.5,
        color="red",
    )

    center_x = (max(volumes) + min(volumes)) / 2
    center_y = (max(energies) + min(energies)) / 2

    # write text of nu value in close middle
    nu = round(nu, 3)
    ax.text(center_x, center_y, f"$\\nu$={nu}")

    ax.legend(loc="upper center", bbox_to_anchor=(0.5, 1.0), fontsize=fontsize)
    ax.set_xlabel("Cell volume per formula unit ($\\AA^3$)", fontsize=fontsize)
    ax.set_ylabel("$E - TS$ (eV)", fontsize=fontsize)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(title, fontsize=fontsize)


@cmd_root.command("inspect")
@click.argument("node")  # The node to inspect, uuid or pk
@click.option("--output", "-o", default="output", help="The output file name")
def inspect(node, output):
    """Inspect the results of the workflow"""
    wf_node = orm.load_node(node)
    if "measure" in wf_node.outputs:
        if "precision" in wf_node.outputs.measure:
            precision = wf_node.outputs.measure.precision

            # print summary of the precision to a json file
            try:
                d_str = json.dumps(precision.output_parameters.get_dict(), indent=4)
                with open(f"{output}_precision_summary.json", "w") as f:
                    f.write(d_str)
            except:
                pass

            # if there are 5 plots, need 3 rows, since the output_parametres is in the dict len(precision) / 2 is the number of rows
            rows = len(precision) // 2

            # create a figure with 2 columns and rows rows on a a4 size paper
            fig, axs = plt.subplots(rows, 2, figsize=(8.27, 11.69), dpi=100)

            # Plot EOS curve and save to a pdf file
            i = 0
            for conf, res in precision.items():
                if conf == "output_parameters":
                    continue

                if res.output_parameters.get_dict() == {}:
                    continue

                # Plot EOS curve
                try:
                    ref_data = res.output_parameters.get_dict()["reference_ae_V0_B0_B1"]
                except KeyError:
                    # For backward compatibility, where I used a different key name before (It is changed after v4.2.0)
                    ref_data = res.output_parameters.get_dict()[
                        "reference_wien2k_V0_B0_B1"
                    ]

                data = res.output_parameters.get_dict()["birch_murnaghan_results"]

                energy0 = res.eos.output_birch_murnaghan_fit.get_dict()["energy0"]
                volumes = res.eos.output_volume_energy.get_dict()["volumes"]
                energies = res.eos.output_volume_energy.get_dict()["energies"]

                nu = res.output_parameters.get_dict()["rel_errors_vec_length"]
                title = f"EOS of {conf}"

                ax = axs.flat[i]
                eos_plot(
                    ax, ref_data, data, energy0, volumes, energies, nu, title=title
                )

                i += 1

            # fig to pdf
            fig.tight_layout()
            fig.savefig(f"{output}_precision.pdf", bbox_inches="tight")


if __name__ == "__main__":
    inspect()



# -*- coding: utf-8 -*-
"""
WorkChain calculate the bands for certain pseudopotential
"""
import numpy as np
from aiida import orm
from aiida.engine import ToContext, calcfunction, if_, while_
from aiida.engine.processes import ExitCode
from aiida.plugins import DataFactory

from . import _BaseEvaluateWorkChain
from ._caching_wise_bands import PwBandsWorkChain

UpfData = DataFactory("pseudo.upf")


@calcfunction
def create_kpoints_from_distance(structure, distance, force_parity):
    """Generate a uniformly spaced kpoint mesh for a given structure.
    The spacing between kpoints in reciprocal space is guaranteed to be at least the defined distance.
    :param structure: the StructureData to which the mesh should apply
    :param distance: a Float with the desired distance between kpoints in reciprocal space
    :param force_parity: a Bool to specify whether the generated mesh should maintain parity
    :returns: a KpointsData with the generated mesh
    """
    from aiida.orm import KpointsData
    from numpy import linalg

    epsilon = 1e-5

    kpoints = KpointsData()
    kpoints.set_cell_from_structure(structure)
    kpoints.set_kpoints_mesh_from_density(
        distance.value, force_parity=force_parity.value
    )

    lengths_vector = [linalg.norm(vector) for vector in structure.cell]
    lengths_kpoint = kpoints.get_kpoints_mesh()[0]

    is_symmetric_cell = all(
        abs(length - lengths_vector[0]) < epsilon for length in lengths_vector
    )
    is_symmetric_mesh = all(length == lengths_kpoint[0] for length in lengths_kpoint)

    # If the vectors of the cell all have the same length, the kpoint mesh should be isotropic as well
    if is_symmetric_cell and not is_symmetric_mesh:
        nkpoints = max(lengths_kpoint)
        kpoints.set_kpoints_mesh([nkpoints, nkpoints, nkpoints])

    return kpoints


def validate_inputs(inputs, ctx=None):
    """Validate the inputs of the entire input namespace"""
    if inputs["run_band_structure"] and (
        "kpoints_distance_band_structure" not in inputs
    ):
        return (
            BandsWorkChain.exit_codes.ERROR_KPOINTS_DISTANCE_BAND_STRUCTURE_NOT_SET.message
        )


class BandsWorkChain(_BaseEvaluateWorkChain):
    """WorkChain calculate the bands for certain pseudopotential
    Can choose only run bands or only on bandstructure"""

    # maximum number of bands factor increase loop
    # to prevent the infinite loop in bands evaluation
    _MAX_NUM_BANDS_FACTOR = 5

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.expose_inputs(PwBandsWorkChain, include=['scf', 'bands', 'structure'])
        spec.input('kpoints_distance_bands', valid_type=orm.Float,
                    help='Kpoints distance setting for bulk energy bands calculation.')
        spec.input('init_nbands_factor', valid_type=orm.Float, default=lambda: orm.Float(1.5),
                    help='initial nbands factor.')
        spec.input('fermi_shift', valid_type=orm.Float, default=lambda: orm.Float(10.0),
                    help='The uplimit of energy to check the bands diff, control the number of bands.')
        spec.input('run_band_structure', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='if True, run bands structure calculation on seekpath kpath.')
        spec.input('kpoints_distance_band_structure', valid_type=orm.Float, required=False,
                    help='Kpoints distance setting for band structure calculation.')
        spec.inputs.validator = validate_inputs

        spec.outline(
            cls.setup,

            while_(cls.should_run_bands)(
                cls.run_bands,
                cls.inspect_bands,
                cls.increase_nbands,
            ),

            if_(cls.should_run_band_structure)(
                cls.run_band_structure,
                cls.inspect_band_structure,
            ),
            cls.finalize,
        )

        spec.expose_outputs(PwBandsWorkChain, namespace='band_structure',
                            namespace_options={'dynamic': True, 'required': False})
        spec.expose_outputs(PwBandsWorkChain, namespace='bands')
        spec.exit_code(201, 'ERROR_SUB_PROCESS_FAILED_BANDS',
                    message='The `PwBandsWorkChain` sub process failed.')
        spec.exit_code(203, 'ERROR_REACH_MAX_BANDS_FACTOR_INCREASE_LOOP',
                    message=f'The maximum number={cls._MAX_NUM_BANDS_FACTOR} of bands factor'
                            'increase loop reached, but still cannot get enough bands.')
        spec.exit_code(204, 'ERROR_KPOINTS_DISTANCE_BAND_STRUCTURE_NOT_SET',
                    message=f'kpoints distance is not set in inputs.')
        # yapf: enable

    def setup(self):
        """Input validation"""
        # set initial lowest highest bands eigenvalue - fermi_energy equals to 0.0
        self.ctx.nbands_factor = self.inputs.init_nbands_factor.value

        # For qe PwBandsWorkChain if `bands_kpoints` not set, the seekpath will run
        # to give a seekpath along the recommonded path.
        # There for I need to explicitly set the bands_kpoints and apply it
        # to the bands workchain.
        # While for the band_structure sub-workchain, the seekpath will run and
        # give band structure along the path.
        self.ctx.bands_kpoints = create_kpoints_from_distance(
            self.inputs.structure, self.inputs.kpoints_distance_bands, orm.Bool(False)
        )

        # break_increase_nbands set to True will force the while_ loop to stop.
        self.ctx.break_increase_nbands = False

        # flag to control whether bands loop keep on running
        self.ctx.should_run_bands = True

    def should_run_bands(self) -> bool:
        return self.ctx.should_run_bands

    def run_bands(self):
        """run bands calculation"""
        inputs = self.exposed_inputs(PwBandsWorkChain)
        inputs["nbands_factor"] = orm.Float(self.ctx.nbands_factor)
        inputs["bands_kpoints"] = self.ctx.bands_kpoints

        running = self.submit(PwBandsWorkChain, **inputs)
        self.report(f"Running pw bands calculation pk={running.pk}")
        return ToContext(workchain_bands=running)

    def inspect_bands(self):
        """
        Return workchain if it is finished_ok
        """
        workchain = self.ctx.workchain_bands

        if workchain.is_finished:
            self._disable_cache(workchain)

        if not workchain.is_finished_ok:
            self.logger.warning(
                f"PwBandsWorkChain for bands evaluation failed with exit status {workchain.exit_status}"
            )

            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_BANDS

        # Check if the bands are enough to cover the desired fermi shift
        fermi_energy = workchain.outputs.band_parameters["fermi_energy"]
        bands = workchain.outputs.band_structure.get_array("bands")

        # -1 colume for the highest eigenvalue of every kpoints
        # which might not be belong to one band if there are degeneracy
        # not enough until eigenvalues of all kpoints are greater than shift value.
        if bands.ndim == 2:
            highest_band = bands[:, -1]
        elif bands.ndim == 3:
            # for magnetization structure, the bands are 3d array.
            highest_band = bands[:, :, -1]

        if np.all(highest_band > fermi_energy + self.inputs.fermi_shift.value):
            self.ctx.should_run_bands = False
        else:
            self.report(
                f"highest band eigenvalue above fermi is {highest_band.min() - fermi_energy}, not enough, you want {self.inputs.fermi_shift.value}."
            )

        self.ctx.ecutwfc = workchain.inputs.scf.pw.parameters["SYSTEM"]["ecutwfc"]
        self.ctx.ecutrho = workchain.inputs.scf.pw.parameters["SYSTEM"]["ecutrho"]

        return workchain

    def increase_nbands(self):
        """inspect the result of bands calculation."""

        # It is here not only because we don't want to stop bands number increase,
        # but also because of a bug that return exit in `while_` not terminate the workflow.
        #
        # If the bands evaluation failed it will keeps on increasing the `nbands_factor`
        # which lead to the infinite loop to this work chain.
        # Here I work around it by giving maximum nbands_factor loop to _MAX_NUM_BANDS_FACTOR=5.
        # And add a `break_increase_nbands` to break and get out from `should_run_bands` immediately.
        if (
            self.ctx.nbands_factor
            > self.inputs.init_nbands_factor.value + self._MAX_NUM_BANDS_FACTOR
        ):
            return self.exit_codes.ERROR_REACH_MAX_BANDS_FACTOR_INCREASE_LOOP

        if self.ctx.should_run_bands:
            # only when bands is not enough, increase the factor and rerun
            # I should not increase it anyway because this ctx is used
            # for the input of band structure calculation.
            self.report(
                f"increasing nbands_factor to from {self.ctx.nbands_factor} to {self.ctx.nbands_factor + 1.0}"
            )
            self.ctx.nbands_factor += 1.0

    def should_run_band_structure(self):
        """whether run band structure"""
        return self.inputs.run_band_structure.value

    def run_band_structure(self):
        """run band structure calculation"""
        inputs = self.exposed_inputs(PwBandsWorkChain)
        inputs["nbands_factor"] = orm.Float(self.ctx.nbands_factor)
        inputs["bands_kpoints_distance"] = orm.Float(
            self.inputs.kpoints_distance_band_structure
        )

        running = self.submit(PwBandsWorkChain, **inputs)
        self.report(f"Running pw band structure calculation pk={running.pk}")
        return ToContext(workchain_band_structure=running)

    def inspect_band_structure(self):
        """inspect band structure and dump all its output"""
        self._inspect_workchain(
            namespace="band_structure", workchain=self.ctx.workchain_band_structure
        )

    def finalize(self):
        """inspect band and dump all its output"""
        self._inspect_workchain(namespace="bands", workchain=self.ctx.workchain_bands)

        self.out("ecutwfc", orm.Int(self.ctx.ecutwfc).store())
        self.out("ecutrho", orm.Int(self.ctx.ecutrho).store())

    def _inspect_workchain(self, namespace, workchain):
        if not workchain.is_finished_ok:
            self.logger.warning(
                f"PwBandsWorkChain uuid={workchain.uuid} failed with exit status {workchain.exit_status}"
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED_BANDS

        self.out_many(
            self.exposed_outputs(workchain, PwBandsWorkChain, namespace=namespace)
        )



# -*- coding: utf-8 -*-
"""Workchain to calculate delta factor of specific psp"""

from aiida import orm
from aiida.engine import if_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import (
    ACTINIDE_ELEMENTS,
    HIGH_DUAL_ELEMENTS,
    LANTHANIDE_ELEMENTS,
    MAGNETIC_ELEMENTS,
    NO_GS_CONF_ELEMENTS,
    OXIDE_CONFIGURATIONS,
    UNARIE_CONFIGURATIONS,
    get_magnetic_inputs,
    get_protocol,
    get_standard_structure,
    reset_pseudos_for_magnetic,
    update_dict,
)
from aiida_sssp_workflow.workflows.common import (
    get_extra_parameters_for_lanthanides,
    get_pseudo_element_and_type,
    get_pseudo_N,
    get_pseudo_O,
)
from aiida_sssp_workflow.workflows.evaluate._metric import MetricWorkChain
from aiida_sssp_workflow.workflows.measure import _BaseMeasureWorkChain
from pseudo_parser.upf_parser import parse_element, parse_pseudo_type

UpfData = DataFactory("pseudo.upf")


class PrecisionMeasureWorkChain(_BaseMeasureWorkChain):
    """Workchain to calculate delta factor of specific pseudopotential"""

    # pylint: disable=too-many-instance-attributes

    _OXIDE_CONFIGURATIONS = OXIDE_CONFIGURATIONS

    #_UNARIE_GS_CONFIGURATIONS = UNARIE_CONFIGURATIONS + ["GS"]
    # For now, we decide not include the GS configuration since the reference data from sci 2016 paper use
    # the different parameters compared with the aiida common workflow. So the result will be different.
    _UNARIE_GS_CONFIGURATIONS = UNARIE_CONFIGURATIONS

    _NBANDS_FACTOR_FOR_LAN = 1.5

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)

        spec.outline(
            cls.setup,
            cls.setup_pw_parameters_from_protocol,
            cls.setup_configurations,
            cls.run_metric,
            cls.inspect_metric,
            cls.finalize,
        )
        # namespace for storing all detail of run on each configuration
        for configuration in cls._OXIDE_CONFIGURATIONS + cls._UNARIE_GS_CONFIGURATIONS + ["RE"]:
            spec.expose_outputs(MetricWorkChain, namespace=configuration,
                        namespace_options={
                            'help': f'Delta calculation result of {configuration} EOS.',
                            'required': False,
                        })

        spec.output('output_parameters',
                    help='The summary output parameters of all delta measures to describe the precision of EOS compare '
                        ' with the AE equation of state.')
        spec.exit_code(401, 'ERROR_METRIC_WORKCHAIN_NOT_FINISHED_OK', message='The metric workchain of configuration {confs} not finished ok.')
        # yapf: enable

    def _setup_pseudo_and_configuration(self):
        """Depend on the element set the proper pseudo and configuration list"""

        # this is the pseudo dict for the element
        self.ctx.pseudos_unary = {self.ctx.element: self.inputs.pseudo}

        # for the oxide, need to pseudo of oxygen,
        # the pseudo is the one select after the oxygen verification and
        # store in the `statics/upf/O.**.upf`
        pseudo_O = get_pseudo_O()
        self.ctx.pseudos_oxide = {
            self.ctx.element: self.inputs.pseudo,
            "O": pseudo_O,
        }

        # For oxygen, still run for oxides but use only the pseudo.
        if self.ctx.element == "O":
            self.ctx.pseudos_oxide = {
                self.ctx.element: self.inputs.pseudo,
            }

        # Structures for delta factor calculation as provided in
        # http:// molmod.ugent.be/deltacodesdft/
        # Exception for lanthanides use nitride structures from
        # https://doi.org/10.1016/j.commatsci.2014.07.030 and from
        # common-workflow set from acwf paper xsf files all store in `statics/structures`.
        # keys here are: BCC, FCC, SC, Diamond, XO, XO2, XO3, X2O, X2O3, X2O5, RE (Lanthanide that will use RE-N), GS
        if self.ctx.element in NO_GS_CONF_ELEMENTS + ACTINIDE_ELEMENTS:
            # Don't have ground state structure for At, Fr, Ra
            self.ctx.configuration_list = (
                self._OXIDE_CONFIGURATIONS + UNARIE_CONFIGURATIONS
            )
        elif self.ctx.element in LANTHANIDE_ELEMENTS:
            self.ctx.configuration_list = (
                self._OXIDE_CONFIGURATIONS + UNARIE_CONFIGURATIONS + ["RE"]
            )
        else:
            self.ctx.configuration_list = (
                self._OXIDE_CONFIGURATIONS + self._UNARIE_GS_CONFIGURATIONS
            )

        # set structures except RARE earth element and actinides elements with will be set independently
        # in sepecific step. Other wise, the gs structure is request but not provided, which
        # will raise error.
        self.ctx.structures = dict()
        for configuration in self.ctx.configuration_list:
            self.ctx.structures[configuration] = get_standard_structure(
                self.ctx.element,
                prop="delta",
                configuration=configuration,
            )

    def setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        content = self.inputs.pseudo.get_content()
        self.ctx.element = parse_element(content)
        self.ctx.pseudo_type = parse_pseudo_type(content)

        self.ctx.pw_parameters = {}

        self._setup_pseudo_and_configuration()

        # set up the ecutwfc and ecutrho
        self.ctx.ecutwfc = self.inputs.wavefunction_cutoff.value
        self.ctx.ecutrho = self.inputs.charge_density_cutoff.value

    def setup_pw_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol = get_protocol(category="precision", name=self.inputs.protocol.value)
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR_PER_ATOM = protocol["conv_thr_per_atom"]
        self._MIXING_BETA = protocol["mixing_beta"]
        self.ctx.kpoints_distance = self._KDISTANCE = protocol["kpoints_distance"]
        self.ctx.scale_count = self._SCALE_COUNT = protocol["scale_count"]
        self.ctx.scale_increment = self._SCALE_INCREMENT = protocol["scale_increment"]

        parameters = {
            "CONTROL": {
                "calculation": "scf",
                "disk_io": "nowf",  # safe to hard-code, this will never be the parent calculation of other calculations
            },
            "SYSTEM": {
                "degauss": self._DEGAUSS,
                "occupations": self._OCCUPATIONS,
                "smearing": self._SMEARING,
            },
            "ELECTRONS": {
                "conv_thr": self._CONV_THR_PER_ATOM,
                "mixing_beta": self._MIXING_BETA,
            },
        }

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, parameters)

        self.logger.info(f"The pw parameters for EOS step is: {self.ctx.pw_parameters}")

    def setup_configurations(self):
        # narrow the configuration list by protocol
        # this is used for test protocol which only has limited configurations to be verified
        if "configurations" in self.inputs:
            clist = self.inputs.configurations.get_list()
        else:
            clist = self.ctx.configuration_list

        for key in list(self.ctx.structures.keys()):
            if key not in clist:
                self.ctx.structures.pop(key)

    @staticmethod
    def _compute_nelectrons_of_oxide(configuration, z_O, z_X):
        """Will return the number of electrons of oxide configurations with pseudos
        z_O is the number of electrons of oxygen pseudo
        z_X is the number of electrons of X pseudo
        """
        if configuration == "XO":
            return z_X + z_O

        elif configuration == "XO2":
            return z_X + z_O * 2

        elif configuration == "XO3":
            return z_X + z_O * 3

        elif configuration == "X2O":
            return z_X * 2 + z_O

        elif configuration == "X2O3":
            return z_X * 4 + z_O * 6

        elif configuration == "X2O5":
            return z_X * 4 + z_O * 10

        else:
            raise ValueError(
                f"Cannot compute the number electrons of configuration {configuration} with z_O={z_O} and z_X={z_X}."
            )

    def _get_inputs(self, structure, configuration):
        """Set pw parameters and pseudos based on configuration and structure"""
        if configuration in self._OXIDE_CONFIGURATIONS:
            # pseudos for oxides
            pseudos = self.ctx.pseudos_oxide
            pw_parameters = self.ctx.pw_parameters
            kpoints_distance = self.ctx.kpoints_distance

            # need also increase nbands for Rare-earth oxides.
            # See https://github.com/aiidateam/aiida-sssp-workflow/issues/161
            # This is not easy to be set in the rare-earth step since it will
            # finally act on here
            if self.ctx.element in LANTHANIDE_ELEMENTS:
                nbnd_factor = self._NBANDS_FACTOR_FOR_LAN
                pseudo_O = get_pseudo_O()
                pseudo_RE = self.inputs.pseudo
                nbnd = (
                    nbnd_factor
                    * int(
                        self._compute_nelectrons_of_oxide(
                            configuration, pseudo_O.z_valence, pseudo_RE.z_valence
                        )
                    )
                    // 2
                )
                pw_parameters["SYSTEM"]["nbnd"] = int(nbnd)

        if configuration in self._UNARIE_GS_CONFIGURATIONS:  # include regular 'GS'
            # pseudos for BCC, FCC, SC, Diamond and TYPYCAL configurations
            pseudos = self.ctx.pseudos_unary
            pw_parameters = self.ctx.pw_parameters
            kpoints_distance = self.ctx.kpoints_distance

        if configuration == "GS" and self.ctx.element in MAGNETIC_ELEMENTS:
            # specific setting for magnetic elements gs since mag on

            # reconstruct configuration, O1, O2 for sites
            (
                structure,
                pw_magnetic_parameters,
            ) = get_magnetic_inputs(structure)

            # override pseudos setting
            # required for O, Mn, Cr where the kind names varies for sites
            self.ctx.pseudos_magnetic = reset_pseudos_for_magnetic(
                self.inputs.pseudo, structure
            )

            pseudos = self.ctx.pseudos_magnetic
            pw_parameters = update_dict(self.ctx.pw_parameters, pw_magnetic_parameters)

        if configuration == "RE":
            # pseudos for nitrides
            pseudo_N = get_pseudo_N()
            pseudo_RE = self.inputs.pseudo
            self.ctx.pseudos_nitride = {"N": pseudo_N, self.ctx.element: pseudo_RE}
            pseudos = self.ctx.pseudos_nitride

            # perticular parameters for RE-N
            # Since the reference data is from https://doi.org/10.1016/j.commatsci.2014.07.030
            # Here I need to use the same input parameters
            nbnd_factor = self._NBANDS_FACTOR_FOR_LAN
            nbnd = nbnd_factor * (pseudo_N.z_valence + pseudo_RE.z_valence)

            pw_parameters = self.ctx.pw_parameters
            # Set the namespace directly will override the original value set in `self.ctx.pw_parameters`
            pw_parameters = update_dict(
                self.ctx.pw_parameters,
                get_extra_parameters_for_lanthanides(self.ctx.element, nbnd),
            )
            pw_parameters["SYSTEM"]["occupations"] = "tetrahedra"
            pw_parameters["SYSTEM"].pop("smearing")

            # sparse kpoints, we use tetrahedra occupation
            kpoints_distance = self.ctx.kpoints_distance + 0.1

        ecutwfc, ecutrho = self._get_pw_cutoff(
            structure, self.ctx.ecutwfc, self.ctx.ecutrho
        )
        parameters = {
            "SYSTEM": {
                "ecutwfc": round(ecutwfc, 1),
                "ecutrho": round(ecutrho, 1),
            },
        }
        parameters = update_dict(parameters, pw_parameters)

        # conv_thr is extensive, like the total energy so we need to scale it with the number of atoms
        natoms = len(structure.sites)
        parameters["ELECTRONS"]["conv_thr"] = (
            parameters["ELECTRONS"]["conv_thr"] * natoms
        )

        inputs = {
            "eos": {
                "metadata": {"call_link_label": "delta_EOS"},
                "structure": structure,
                "kpoints_distance": orm.Float(kpoints_distance),
                "scale_count": orm.Int(self.ctx.scale_count),
                "scale_increment": orm.Float(self.ctx.scale_increment),
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": pseudos,
                    "parameters": orm.Dict(dict=parameters),
                    "metadata": {
                        "options": self.inputs.options.get_dict(),
                    },
                    "parallelization": self.inputs.parallelization,
                },
            },
            "element": orm.Str(self.ctx.element),
            "configuration": orm.Str(configuration),
            "clean_workdir": self.inputs.clean_workdir,
        }

        return inputs

    def run_metric(self):
        """run eos workchain"""

        for configuration, structure in self.ctx.structures.items():
            inputs = self._get_inputs(structure, configuration)

            future = self.submit(MetricWorkChain, **inputs)
            self.report(
                f"launching DeltaWarkChain<{future.pk}> for {configuration} structure."
            )

            self.to_context(**{f"{configuration}_metric": future})

    def inspect_metric(self):
        """Inspect the results of MetricWorkChain"""
        failed_configuration_lst = list()
        for configuration in self.ctx.structures.keys():
            workchain = self.ctx[f"{configuration}_metric"]

            if not workchain.is_finished_ok:
                self.logger.warning(
                    f"MetricWorkChain of {configuration} failed with exit status {workchain.exit_status}"
                )
                failed_configuration_lst.append(configuration)

            self.out_many(
                self.exposed_outputs(
                    workchain,
                    MetricWorkChain,
                    namespace=configuration,
                )
            )

        if failed_configuration_lst:
            return self.exit_codes.ERROR_METRIC_WORKCHAIN_NOT_FINISHED_OK.format(
                confs=f"{failed_configuration_lst}",
            )

    def finalize(self):
        """calculate the delta factor"""
        output_parameters = {}

        for configuration in self.ctx.configuration_list:
            try:
                output = self.outputs[configuration].get("output_parameters")
            except KeyError:
                self.logger.warning(
                    f"Can not get the key {configuration} from outputs, not verify or failed."
                )
                continue

            try:
                output_parameters[configuration] = {
                    "delta": output["delta"],
                    "delta/natoms": output["delta/natoms"],
                    "nu": output["rel_errors_vec_length"],
                }
            except KeyError:
                self.logger.warning(
                    f"Can not get the metric, check EOS result or directly recalculate metric from EOS."
                )
                continue

        self.out("output_parameters", orm.Dict(dict=output_parameters).store())



import importlib
from typing import Optional

from aiida import orm
from aiida.plugins import DataFactory

from pseudo_parser.upf_parser import parse_element, parse_pseudo_type

UpfData = DataFactory("pseudo.upf")


def get_pseudo_element_and_type(pseudo):
    """Giving a pseudo, return element and pseudo type as tuple"""
    content = pseudo.get_content()
    element = parse_element(content)
    pseudo_type = parse_pseudo_type(content)

    return element, pseudo_type


def get_extra_parameters_for_lanthanides(element, nbnd) -> dict:
    """
    In rare earth case, increase the initial number of bands,
    otherwise the occupation will not fill up in the highest band
    which always trigger the `PwBaseWorkChain` sanity check.
    The nbnd only take effect for the scf step of PwBandsWorkChain
    since for the bands step, the nbnd is controled by init_nbands_factor
    while `nbnd` will be removed from scf parameters.
    """
    extra_parameters = {
        "SYSTEM": {
            "nspin": 2,
            "starting_magnetization": {
                element: 0.5,
            },
            "nbnd": int(nbnd),
        },
        "ELECTRONS": {
            "diagonalization": "cg",
            "mixing_beta": 0.5,
            "electron_maxstep": 200,
        },
    }

    return extra_parameters


def get_pseudo_N():
    """Return pseudo of nitrogen for lanthanide nitrides"""
    import_path = importlib.resources.path(
        "aiida_sssp_workflow.statics.upf", "N.us.z_5.ld1.psl.v0.1.upf"
    )
    with import_path as psp_path, open(psp_path, "rb") as stream:
        pseudo_N = UpfData(stream)

    return pseudo_N


def get_pseudo_O():
    """Return pseudo of oxygen for oxides"""
    import_path = importlib.resources.path(
        "aiida_sssp_workflow.statics.upf", "O.paw.z_6.ld1.psl.v0.1.upf"
    )
    with import_path as psp_path, open(psp_path, "rb") as stream:
        pseudo_O = UpfData(stream)

    return pseudo_O


def clean_workdir(node: orm.CalcJobNode) -> Optional[int]:
    """clean remote workdir of nonmenon calcjob"""
    # I have to do only clean nonmenon calcjob since I regard it as a bug that
    # the workdir of cached node point to the identical remote path of nomenon calcjob node.
    # It is like a soft link or shallow copy in caching. This lead to clean the remote path
    # of cached node also destroy the remote path of nonmenon node that may still be used for
    # other subsequent calcjob as `parent_folder`, i.e PH calculation.
    if "_aiida_cached_from" not in node.extras:
        node.outputs.remote_folder._clean()  # pylint: disable=protected-access
        return node.pk
    else:
        return None


def invalid_cache(node: orm.CalcJobNode) -> Optional[int]:
    """invalid cache of cached calcjob, so it will not be used for further caching"""

    if node.is_valid_cache:
        node.is_valid_cache = False
        return node.pk
    else:
        return None


def operate_calcjobs(wnode, operator, all_same_nodes=False):
    """iterate over desendant calcjob nodes of given workflow node and apply operator.

    :param wnode: workflow node to operating on its descendat calcjob nodes
    :param operator: a function operate to the descendants calcjob node
    :param all_same_nodes: If `True`, fetch all same node and process the operation. BE CAREFUL!
    :return: return the list of nodes pk being cleaned
    """

    cleaned_calcs = []
    for child in wnode.called_descendants:
        if isinstance(child, orm.CalcJobNode):
            # the nodes waid for operated.
            nodes = child.get_all_same_nodes()
            try:
                if not all_same_nodes:
                    # only operated on this node
                    nodes = [child]

                for n in nodes:
                    pk = operator(n)
                    cleaned_calcs += [pk] if pk else []

            except (IOError, OSError, KeyError) as exc:
                raise RuntimeError(
                    "Failed to clean working dirctory of calcjob"
                ) from exc

    return cleaned_calcs



"""Base Measure Workchain since bands measure and precision measure share same input ports
This class makes it able to expose precision measure and bands measure inputs to verification
workchain.
"""
from aiida import orm
from aiida.plugins import DataFactory

from aiida_sssp_workflow.workflows import SelfCleanWorkChain

UpfData = DataFactory("pseudo.upf")


class _BaseMeasureWorkChain(SelfCleanWorkChain):
    """Base Measure Workchain since bands measure and precision measure share same input ports"""

    # ECUT for oxygen, remember to update this if the oxygen pseudo is changed
    # Currently the oxygen pseudo is `O.paw.z_6.ld1.psl.v0.1.upf`
    _O_ECUTWFC = 75.0
    _O_ECUTRHO = 560.0

    # ECUT for nitrogen, remember to update this if the nitrogen pseudo is changed
    # Currently the nitrogen pseudo is `N.us.z_5.ld1.psl.v0.1.upf`
    _N_ECUTWFC = 55.0
    _N_ECUTRHO = 330.0

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('code', valid_type=orm.AbstractCode,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input('pseudo', valid_type=UpfData, required=True,
                    help='Pseudopotential to be verified')
        spec.input('protocol', valid_type=orm.Str, required=True,
                    help='The protocol which define input calculation parameters.')
        spec.input('wavefunction_cutoff', valid_type=orm.Float, required=True, help='The wavefunction cutoff.')
        spec.input('configurations', valid_type=orm.List, required=False)
        spec.input('charge_density_cutoff', valid_type=orm.Float, required=True, help='The charge density cutoff.')
        spec.input('options', valid_type=orm.Dict, required=True,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=True,
                    help='Parallelization options for the `PwCalculations`.')

    def _get_pw_cutoff(
        self, structure: orm.StructureData, ecutwfc: float, ecutrho: float
    ):
        """Get cutoff pair, if strcture contains oxygen or nitrogen, need
        to use the max between pseudo cutoff and the O/N cutoff.
        """
        elements = set(structure.get_symbols_set())
        if "O" in elements:
            ecutwfc = max(ecutwfc, self._O_ECUTWFC)
            ecutrho = max(ecutrho, self._O_ECUTRHO)

        if "N" in elements:
            ecutwfc = max(ecutwfc, self._N_ECUTWFC)
            ecutrho = max(ecutrho, self._N_ECUTRHO)

        return ecutwfc, ecutrho



# -*- coding: utf-8 -*-
"""
Bands distance of many input pseudos
"""

from aiida import orm
from aiida.engine import ToContext, if_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import (
    LANTHANIDE_ELEMENTS,
    MAGNETIC_ELEMENTS,
    NONMETAL_ELEMENTS,
    get_magnetic_inputs,
    get_protocol,
    get_standard_structure,
    reset_pseudos_for_magnetic,
    update_dict,
)
from aiida_sssp_workflow.workflows.common import (
    get_extra_parameters_for_lanthanides,
    get_pseudo_N,
)
from aiida_sssp_workflow.workflows.evaluate._bands import BandsWorkChain
from aiida_sssp_workflow.workflows.measure import _BaseMeasureWorkChain

UpfData = DataFactory("pseudo.upf")


def validate_input_pseudos(d_pseudos, _):
    """Validate that all input pseudos map to same element"""
    element = set(pseudo.element for pseudo in d_pseudos.values())

    if len(element) > 1:
        return f"The pseudos corespond to different elements {element}."


class BandsMeasureWorkChain(_BaseMeasureWorkChain):
    """
    WorkChain to run bands measure,
    run without sym for distance compare and band structure along the path
    """

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)

        spec.outline(
            cls.setup,
            if_(cls.is_magnetic_element)(
                cls.extra_setup_for_magnetic_element,
            ),
            if_(cls.is_lanthanide_element)(
                cls.extra_setup_for_lanthanide_element, ),
            cls.setup_pw_parameters_from_protocol,
            cls.run_bands_evaluation,
            cls.finalize,
        )

        spec.expose_outputs(BandsWorkChain)

    def setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        self.ctx.pw_parameters = {}

        element = self.ctx.element = self.inputs.pseudo.element
        self.ctx.pseudos = {self.ctx.element: self.inputs.pseudo}

        self.ctx.structure = get_standard_structure(
            element,
            prop="bands",
        )

        # extra setting for bands convergence
        self.ctx.is_metal = element not in NONMETAL_ELEMENTS

        # set up the ecutwfc and ecutrho
        self.ctx.ecutwfc = self.inputs.wavefunction_cutoff.value
        self.ctx.ecutrho = self.inputs.charge_density_cutoff.value

    def is_magnetic_element(self):
        """Check if the element is magnetic"""
        return self.ctx.element in MAGNETIC_ELEMENTS

    def extra_setup_for_magnetic_element(self):
        """Extra setup for magnetic element"""
        self.ctx.structure, magnetic_extra_parameters = get_magnetic_inputs(
            self.ctx.structure
        )
        self.ctx.pw_parameters = update_dict(
            self.ctx.pw_parameters, magnetic_extra_parameters
        )

        # override pseudos setting
        # required for O, Mn, Cr where the kind names varies for sites
        self.ctx.pseudos = reset_pseudos_for_magnetic(
            self.inputs.pseudo, self.ctx.structure
        )

    def is_lanthanide_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in LANTHANIDE_ELEMENTS

    def extra_setup_for_lanthanide_element(self):
        """Extra setup for rare earth element"""
        nbnd_factor = 2.0
        pseudo_N = get_pseudo_N()
        self.ctx.pseudos["N"] = pseudo_N
        pseudo_RE = self.inputs.pseudo
        nbnd = nbnd_factor * (pseudo_N.z_valence + pseudo_RE.z_valence)
        pw_parameters = get_extra_parameters_for_lanthanides(self.ctx.element, nbnd)

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, pw_parameters)

    def setup_pw_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol = get_protocol(category="bands", name=self.inputs.protocol.value)
        self._DEGAUSS = protocol["degauss"]
        self._OCCUPATIONS = protocol["occupations"]
        self._SMEARING = protocol["smearing"]
        self._CONV_THR_PER_ATOM = protocol["conv_thr_per_atom"]

        self._INIT_NBANDS_FACTOR = protocol["init_nbands_factor"]
        self._FERMI_SHIFT = protocol["fermi_shift"]

        self.ctx.init_nbands_factor = self._INIT_NBANDS_FACTOR
        self.ctx.fermi_shift = self._FERMI_SHIFT

        self.ctx.kpoints_distance_scf = protocol["kpoints_distance_scf"]
        self.ctx.kpoints_distance_bands = protocol["kpoints_distance_bands"]
        self.ctx.kpoints_distance_band_structure = protocol[
            "kpoints_distance_band_structure"
        ]

        parameters = {
            "SYSTEM": {
                "degauss": self._DEGAUSS,
                "occupations": self._OCCUPATIONS,
                "smearing": self._SMEARING,
            },
            "ELECTRONS": {
                "conv_thr": self._CONV_THR_PER_ATOM,
            },
            "CONTROL": {
                "calculation": "scf",
            },
        }

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters, parameters)

        self.logger.info(
            f"The pw parameters for convergence is: {self.ctx.pw_parameters}"
        )

    def _get_inputs(self, pseudos):
        """
        get inputs for the bands evaluation with given pseudo
        """
        ecutwfc, ecutrho = self._get_pw_cutoff(
            self.ctx.structure, self.ctx.ecutwfc, self.ctx.ecutrho
        )

        parameters = {
            "SYSTEM": {
                "ecutwfc": round(ecutwfc, 1),
                "ecutrho": round(ecutrho, 1),
            },
        }
        parameters = update_dict(parameters, self.ctx.pw_parameters)

        parameters_bands = update_dict(parameters, {})
        parameters_bands["SYSTEM"].pop("nbnd", None)
        parameters_bands["CONTROL"]["calculation"] = "bands"

        inputs = {
            "structure": self.ctx.structure,
            "scf": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": pseudos,
                    "parameters": orm.Dict(dict=parameters),
                    "metadata": {
                        "options": self.inputs.options.get_dict(),
                    },
                    "parallelization": self.inputs.parallelization,
                },
                "kpoints_distance": orm.Float(self.ctx.kpoints_distance_scf),
            },
            "bands": {
                "pw": {
                    "code": self.inputs.code,
                    "pseudos": pseudos,
                    "parameters": orm.Dict(dict=parameters_bands),
                    "metadata": {
                        "options": self.inputs.options.get_dict(),
                    },
                    "parallelization": self.inputs.parallelization,
                },
            },
            "kpoints_distance_bands": orm.Float(self.ctx.kpoints_distance_bands),
            "init_nbands_factor": orm.Float(self.ctx.init_nbands_factor),
            "fermi_shift": orm.Float(self.ctx.fermi_shift),
            "run_bands_structure": orm.Bool(True),
            "kpoints_distance_band_structure": orm.Float(
                self.ctx.kpoints_distance_band_structure
            ),
            "clean_workdir": self.inputs.clean_workdir,
        }

        return inputs

    def run_bands_evaluation(self):
        """run bands evaluation of psp in inputs list"""
        inputs = self._get_inputs(self.ctx.pseudos)

        running = self.submit(BandsWorkChain, **inputs)

        self.report(f"launching BandsWorkChain<{running.pk}>")

        return ToContext(bands=running)

    def finalize(self):
        """inspect bands run results"""
        self.out_many(self.exposed_outputs(self.ctx.bands, BandsWorkChain))



from ._base import _BaseConvergenceWorkChain
from .bands import ConvergenceBandsWorkChain
from .caching import _CachingConvergenceWorkChain
from .cohesive_energy import ConvergenceCohesiveEnergyWorkChain
from .delta import ConvergenceDeltaWorkChain
from .phonon_frequencies import ConvergencePhononFrequenciesWorkChain
from .pressure import ConvergencePressureWorkChain

__all__ = (
    "_BaseConvergenceWorkChain",
    "_CachingConvergenceWorkChain",
    "ConvergenceDeltaWorkChain",
    "ConvergenceCohesiveEnergyWorkChain",
    "ConvergencePhononFrequenciesWorkChain",
    "ConvergencePressureWorkChain",
    "ConvergenceBandsWorkChain",
)
